/*******************************************************************************
  MPLAB Harmony Application Header File

  Company:
    Microchip Technology Inc.

  File Name:
    app_fs_tcp.h

  Summary:
    This header file provides prototypes and definitions for the application.

  Description:
    This header file provides function prototypes and data type definitions for
    the application.  Some of these are required by the system (such as the
    "APP_FS_Initialize" and "APP_FS_Tasks" prototypes) and some of them are only used
    internally by the application (such as the "APP_FS_STATES" definition).  Both
    are defined here for convenience.
*******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
* Copyright (C) 2018 Microchip Technology Inc. and its subsidiaries.
*
* Subject to your compliance with these terms, you may use Microchip software
* and any derivatives exclusively with Microchip products. It is your
* responsibility to comply with third party license terms applicable to your
* use of third party software (including open source software) that may
* accompany Microchip software.
*
* THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
* EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
* WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
* PARTICULAR PURPOSE.
*
* IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE,
* INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND
* WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS
* BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE. TO THE
* FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN
* ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY,
* THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
*******************************************************************************/
//DOM-IGNORE-END

#ifndef APP_FS_H
#define APP_FS_H

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************


#include <stdbool.h>
#include <stddef.h>

#include "configuration.h"
#include "definitions.h"

// DOM-IGNORE-BEGIN
#ifdef __cplusplus  // Provide C++ Compatibility

extern "C" {

#endif
// DOM-IGNORE-END


#ifdef __SAM9X60__
#define LED_AH_PB11_Get() (PIOB_REGS->PIO_PDSR & (1<<11))
#define LED_AH_PB12_Get() (PIOB_REGS->PIO_PDSR & (1<<12))
#define LED_AH_PB13_Get() (PIOB_REGS->PIO_PDSR & (1<<13))

#define APP_LED_2StateGet()         LED_AH_PB12_Get()
#define APP_LED_3StateGet()         LED_AH_PB13_Get()

#define APP_SWITCH_1StateGet()      (0)
#define APP_LED_1StateGet()         LED_AH_PB11_Get()
#define APP_LED_1StateSet()         LED_AH_PB11_On()
#define APP_LED_1StateClear()       LED_AH_PB11_Off()
#define APP_LED_1StateToggle()      LED_AH_PB11_Toggle()
#else
#if defined (__PIC32C__) || defined(__SAMA5D2__)
/* This section is highly customizable based on application's specific needs. */
#define APP_SWITCH_1StateGet()      SWITCH_Get()
#define APP_SWITCH_2StateGet()      SWITCH_Get()
#define APP_SWITCH_3StateGet()      SWITCH_Get()
#elif defined (__PIC32MZ__)
/* This section is highly customizable based on application's specific needs. */
#define APP_SWITCH_1StateGet()      SWITCH1_Get()
#define APP_SWITCH_2StateGet()      SWITCH2_Get()
#define APP_SWITCH_3StateGet()      SWITCH2_Get()
#endif

#define APP_LED_1StateSet()         LED1_On()
#define APP_LED_1StateGet()         LED1_Get()
#define APP_LED_1StateClear()       LED1_Off()
#define APP_LED_1StateToggle()      LED1_Toggle()

#if defined(__PIC32MZ__)
#define APP_LED_2StateSet()         LED2_On()
#define APP_LED_2StateGet()         LED2_Get()
#define APP_LED_2StateClear()       LED2_Off()
#define APP_LED_2StateToggle()      LED2_Toggle()

#define APP_LED_3StateSet()         LED3_On()
#define APP_LED_3StateGet()         LED3_Get()
#define APP_LED_3StateClear()       LED3_Off()
#define APP_LED_3StateToggle()      LED3_Toggle()
#endif
#endif


// *****************************************************************************
// *****************************************************************************
// Section: Type Definitions
// *****************************************************************************
// *****************************************************************************

#define BUF_SIZE    (20 * 1024)
#define SDCARD_MOUNT_NAME    "/mnt/mchpSite1/web_pages"//"/mnt/mchpSite1"//"/mnt/mydrive"
#define SDCARD_DEV_NAME      "/dev/mmcblka1"
#define SDCARD_DIR_NAME      "web_pages/arducam"

#define MAX_FILE_NAMES                          255
#define MAX_FILE_NAME_LEN                       255
#define JPEG_FILE_NAME                          24

// *****************************************************************************
/* Application states

  Summary:
    Application states enumeration

  Description:
    This enumeration defines the valid application states.  These states
    determine the behavior of the application at various times.
*/

typedef enum
{
    /* Application's state machine's initial state. */
    APP_FS_TCPIP_STATE_MOUNT_DISK = 0,

    /* The app unmounts the disk */
    APP_FS_TCPIP_STATE_UNMOUNT_DISK,

    /* The app mounts the disk again */
    APP_FS_TCPIP_STATE_MOUNT_DISK_AGAIN,

    /* Set the current drive */
    APP_FS_TCPIP_STATE_SET_CURRENT_DRIVE,

    /* The app closes the file and idles */
    APP_FS_TCPIP_STATE_IDLE,

    /* An app error has occurred */
    APP_FS_TCPIP_STATE_ERROR,

    /* In this state, the application waits for the initialization of the TCP/IP stack
       to complete. */
    APP_FS_TCPIP_STATE_WAIT_INIT,

    /* In this state, the application can do TCP/IP transactions. */
    APP_FS_TCPIP_STATE_TRANSACT,

} APP_FS_TCPIP_STATES;


// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    Application strings and buffers are be defined outside this structure.
 */

typedef struct
{
    /* Application's current state */
    APP_FS_TCPIP_STATES                  state;

    bool fs_sys_ready;

} APP_FS_TCPIP_DATA;


// *****************************************************************************
/* LED State

  Summary:
    Enumerates the supported LED states.

  Description:
    This enumeration defines the supported LED states.

  Remarks:
    None.
*/
typedef enum
{
    /* LED State is on */
    APP_FS_TCPIP_LED_STATE_OFF = 0,
   /* LED State is off */
    APP_FS_TCPIP_LED_STATE_ON = 1,

} APP_FS_TCPIP_LED_STATE;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Routines
// *****************************************************************************
// *****************************************************************************
/* These routines are called by drivers when certain events occur.
*/

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_FS_Initialize ( void )

  Summary:
     MPLAB Harmony application initialization routine.

  Description:
    This function initializes the Harmony application.  It places the
    application in its initial state and prepares it to run so that its
    APP_FS_Tasks function can be called.

  Precondition:
    All other system initialization routines should be called before calling
    this routine (in "SYS_Initialize").

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    APP_FS_Initialize();
    </code>

  Remarks:
    This routine must be called from the SYS_Initialize function.
*/

void APP_FS_TCPIP_Initialize ( void );
bool APP_FS_SystemIsReady(void);

/*******************************************************************************
  Function:
    void APP_FS_Tasks ( void )

  Summary:
    MPLAB Harmony Demo application tasks function

  Description:
    This routine is the Harmony Demo application's tasks function.  It
    defines the application's state machine and core logic.

  Precondition:
    The system and application initialization ("SYS_Initialize") should be
    called before calling this.

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    APP_FS_Tasks();
    </code>

  Remarks:
    This routine must be called from SYS_Tasks() routine.
 */

void APP_FS_TCPIP_Tasks( void );

// HTTP application processing
#if defined(TCPIP_STACK_USE_HTTP_NET_SERVER)
extern void HTTP_APP_Initialize(void);
#endif // defined(TCPIP_STACK_USE_HTTP_NET_SERVER)


#endif /* APP_FS_H */

//DOM-IGNORE-BEGIN
#ifdef __cplusplus
}
#endif
//DOM-IGNORE-END

/*******************************************************************************
 End of File
 */

